@@ -21,6 +21,12 @@
 # IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 # CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+"""Keycloak OpenID module.
+
+The module contains mainly the implementation of KeycloakOpenID class, the main
+class to handle authentication and token manipulation.
+"""
+
 import json
 
 from jose import jwt
@@ -52,8 +58,7 @@ from .urls_patterns import (
 
 
 class KeycloakOpenID:
-    """
-    Keycloak OpenID client.
+    """Keycloak OpenID client.
 
     :param server_url: Keycloak server url
     :param client_id: client id
@@ -62,6 +67,7 @@ class KeycloakOpenID:
     :param verify: True if want check connection SSL
     :param custom_headers: dict of custom header to pass to each HTML request
     :param proxies: dict of proxies to sent the request by.
+    :param timeout: connection timeout in seconds
     """
 
     def __init__(
@@ -75,6 +81,25 @@ class KeycloakOpenID:
         proxies=None,
         timeout=60,
     ):
+        """Init method.
+
+        :param server_url: Keycloak server url
+        :type server_url: str
+        :param client_id: client id
+        :type client_id: str
+        :param realm_name: realm name
+        :type realm_name: str
+        :param client_secret_key: client secret key
+        :type client_secret_key: str
+        :param verify: True if want check connection SSL
+        :type verify: bool
+        :param custom_headers: dict of custom header to pass to each HTML request
+        :type custom_headers: dict
+        :param proxies: dict of proxies to sent the request by.
+        :type proxies: dict
+        :param timeout: connection timeout in seconds
+        :type timeout: int
+        """
         self.client_id = client_id
         self.client_secret_key = client_secret_key
         self.realm_name = realm_name
@@ -87,6 +112,11 @@ class KeycloakOpenID:
 
     @property
     def client_id(self):
+        """Get client id.
+
+        :returns: Client id
+        :rtype: str
+        """
         return self._client_id
 
     @client_id.setter
@@ -95,6 +125,11 @@ class KeycloakOpenID:
 
     @property
     def client_secret_key(self):
+        """Get the client secret key.
+
+        :returns: Client secret key
+        :rtype: str
+        """
         return self._client_secret_key
 
     @client_secret_key.setter
@@ -103,6 +138,11 @@ class KeycloakOpenID:
 
     @property
     def realm_name(self):
+        """Get the realm name.
+
+        :returns: Realm name
+        :rtype: str
+        """
         return self._realm_name
 
     @realm_name.setter
@@ -111,6 +151,11 @@ class KeycloakOpenID:
 
     @property
     def connection(self):
+        """Get connection.
+
+        :returns: Connection manager object
+        :rtype: ConnectionManager
+        """
         return self._connection
 
     @connection.setter
@@ -119,6 +164,11 @@ class KeycloakOpenID:
 
     @property
     def authorization(self):
+        """Get authorization.
+
+        :returns: The authorization manager
+        :rtype: Authorization
+        """
         return self._authorization
 
     @authorization.setter
@@ -126,11 +176,12 @@ class KeycloakOpenID:
         self._authorization = value
 
     def _add_secret_key(self, payload):
-        """
-        Add secret key if exist.
+        """Add secret key if exists.
 
-        :param payload:
-        :return:
+        :param payload: Payload
+        :type payload: dict
+        :returns: Payload with the secret key
+        :rtype: dict
         """
         if self.client_secret_key:
             payload.update({"client_secret": self.client_secret_key})
@@ -138,20 +189,26 @@ class KeycloakOpenID:
         return payload
 
     def _build_name_role(self, role):
-        """
+        """Build name of a role.
 
-        :param role:
-        :return:
+        :param role: Role name
+        :type role: str
+        :returns: Role path
+        :rtype: str
         """
         return self.client_id + "/" + role
 
     def _token_info(self, token, method_token_info, **kwargs):
-        """
+        """Getter for the token data.
 
-        :param token:
-        :param method_token_info:
-        :param kwargs:
-        :return:
+        :param token: Token
+        :type token: str
+        :param method_token_info: Token info method to use
+        :type method_token_info: str
+        :param kwargs: Additional keyword arguments
+        :type kwargs: dict
+        :returns: Token info
+        :rtype: dict
         """
         if method_token_info == "introspect":
             token_info = self.introspect(token)
@@ -161,29 +218,37 @@ class KeycloakOpenID:
         return token_info
 
     def well_known(self):
-        """The most important endpoint to understand is the well-known configuration
+        """Get the well_known object.
+
+        The most important endpoint to understand is the well-known configuration
         endpoint. It lists endpoints and other configuration options relevant to
         the OpenID Connect implementation in Keycloak.
 
-        :return It lists endpoints and other configuration options relevant.
+        :returns: It lists endpoints and other configuration options relevant
+        :rtype: dict
         """
-
         params_path = {"realm-name": self.realm_name}
         data_raw = self.connection.raw_get(URL_WELL_KNOWN.format(**params_path))
-
         return raise_error_from_response(data_raw, KeycloakGetError)
 
-    def auth_url(self, redirect_uri):
-        """
+    def auth_url(self, redirect_uri, scope="email", state=""):
+        """Get authorization URL endpoint.
 
-        http://openid.net/specs/openid-connect-core-1_0.html#AuthorizationEndpoint
-
-        :return:
+        :param redirect_uri: Redirect url to receive oauth code
+        :type redirect_uri: str
+        :param scope: Scope of authorization request, split with the blank space
+        :type scope: str
+        :param state: State will be returned to the redirect_uri
+        :type state: str
+        :returns: Authorization URL Full Build
+        :rtype: str
         """
         params_path = {
             "authorization-endpoint": self.well_known()["authorization_endpoint"],
             "client-id": self.client_id,
             "redirect-uri": redirect_uri,
+            "scope": scope,
+            "state": state,
         }
         return URL_AUTH.format(**params_path)
 
@@ -195,9 +260,11 @@ class KeycloakOpenID:
         code="",
         redirect_uri="",
         totp=None,
+        scope="openid",
         **extra
     ):
-        """
+        """Retrieve user token.
+
         The token endpoint is used to obtain tokens. Tokens can either be obtained by
         exchanging an authorization code or by supplying credentials directly depending on
         what flow is used. The token endpoint is also used to obtain new access tokens
@@ -205,13 +272,24 @@ class KeycloakOpenID:
 
         http://openid.net/specs/openid-connect-core-1_0.html#TokenEndpoint
 
-        :param username:
-        :param password:
-        :param grant_type:
-        :param code:
-        :param redirect_uri:
-        :param totp:
-        :return:
+        :param username: Username
+        :type username: str
+        :param password: Password
+        :type password: str
+        :param grant_type: Grant type
+        :type grant_type: str
+        :param code: Code
+        :type code: str
+        :param redirect_uri: Redirect URI
+        :type redirect_uri: str
+        :param totp: Time-based one-time password
+        :type totp: int
+        :param scope: Scope, defaults to openid
+        :type scope: str
+        :param extra: Additional extra arguments
+        :type extra: dict
+        :returns: Keycloak token
+        :rtype: dict
         """
         params_path = {"realm-name": self.realm_name}
         payload = {
@@ -221,6 +299,7 @@ class KeycloakOpenID:
             "grant_type": grant_type,
             "code": code,
             "redirect_uri": redirect_uri,
+            "scope": scope,
         }
         if extra:
             payload.update(extra)
@@ -230,10 +309,11 @@ class KeycloakOpenID:
 
         payload = self._add_secret_key(payload)
         data_raw = self.connection.raw_post(URL_TOKEN.format(**params_path), data=payload)
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakPostError)
 
     def refresh_token(self, refresh_token, grant_type=["refresh_token"]):
-        """
+        """Refresh the user token.
+
         The token endpoint is used to obtain tokens. Tokens can either be obtained by
         exchanging an authorization code or by supplying credentials directly depending on
         what flow is used. The token endpoint is also used to obtain new access tokens
@@ -241,9 +321,12 @@ class KeycloakOpenID:
 
         http://openid.net/specs/openid-connect-core-1_0.html#TokenEndpoint
 
-        :param refresh_token:
-        :param grant_type:
-        :return:
+        :param refresh_token: Refresh token from Keycloak
+        :type refresh_token: str
+        :param grant_type: Grant type
+        :type grant_type: str
+        :returns: New token
+        :rtype: dict
         """
         params_path = {"realm-name": self.realm_name}
         payload = {
@@ -253,97 +336,126 @@ class KeycloakOpenID:
         }
         payload = self._add_secret_key(payload)
         data_raw = self.connection.raw_post(URL_TOKEN.format(**params_path), data=payload)
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakPostError)
+
+    def exchange_token(
+        self,
+        token: str,
+        client_id: str,
+        audience: str,
+        subject: str,
+        requested_token_type: str = "urn:ietf:params:oauth:token-type:refresh_token",
+        scope: str = "openid",
+    ) -> dict:
+        """Exchange user token.
 
-    def exchange_token(self, token: str, client_id: str, audience: str, subject: str) -> dict:
-        """
         Use a token to obtain an entirely different token. See
         https://www.keycloak.org/docs/latest/securing_apps/index.html#_token-exchange
 
-        :param token:
-        :param client_id:
-        :param audience:
-        :param subject:
-        :return:
+        :param token: Access token
+        :type token: str
+        :param client_id: Client id
+        :type client_id: str
+        :param audience: Audience
+        :type audience: str
+        :param subject: Subject
+        :type subject: str
+        :param requested_token_type: Token type specification
+        :type requested_token_type: str
+        :param scope: Scope, defaults to openid
+        :type scope: str
+        :returns: Exchanged token
+        :rtype: dict
         """
         params_path = {"realm-name": self.realm_name}
         payload = {
             "grant_type": ["urn:ietf:params:oauth:grant-type:token-exchange"],
             "client_id": client_id,
             "subject_token": token,
-            "requested_token_type": "urn:ietf:params:oauth:token-type:refresh_token",
+            "requested_token_type": requested_token_type,
             "audience": audience,
             "requested_subject": subject,
+            "scope": scope,
         }
         payload = self._add_secret_key(payload)
         data_raw = self.connection.raw_post(URL_TOKEN.format(**params_path), data=payload)
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakPostError)
 
     def userinfo(self, token):
-        """
+        """Get the user info object.
+
         The userinfo endpoint returns standard claims about the authenticated user,
         and is protected by a bearer token.
 
         http://openid.net/specs/openid-connect-core-1_0.html#UserInfo
 
-        :param token:
-        :return:
+        :param token: Access token
+        :type token: str
+        :returns: Userinfo object
+        :rtype: dict
         """
-
         self.connection.add_param_headers("Authorization", "Bearer " + token)
         params_path = {"realm-name": self.realm_name}
-
         data_raw = self.connection.raw_get(URL_USERINFO.format(**params_path))
-
         return raise_error_from_response(data_raw, KeycloakGetError)
 
     def logout(self, refresh_token):
-        """
-        The logout endpoint logs out the authenticated user.
-        :param refresh_token:
-        :return:
+        """Log out the authenticated user.
+
+        :param refresh_token: Refresh token from Keycloak
+        :type refresh_token: str
+        :returns: Keycloak server response
+        :rtype: dict
         """
         params_path = {"realm-name": self.realm_name}
         payload = {"client_id": self.client_id, "refresh_token": refresh_token}
-
         payload = self._add_secret_key(payload)
         data_raw = self.connection.raw_post(URL_LOGOUT.format(**params_path), data=payload)
-
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPostError, expected_codes=[204])
 
     def certs(self):
-        """
+        """Get certificates.
+
         The certificate endpoint returns the public keys enabled by the realm, encoded as a
         JSON Web Key (JWK). Depending on the realm settings there can be one or more keys enabled
         for verifying tokens.
 
         https://tools.ietf.org/html/rfc7517
 
-        :return:
+        :returns: Certificates
+        :rtype: dict
         """
         params_path = {"realm-name": self.realm_name}
         data_raw = self.connection.raw_get(URL_CERTS.format(**params_path))
         return raise_error_from_response(data_raw, KeycloakGetError)
 
     def public_key(self):
-        """
+        """Retrieve the public key.
+
         The public key is exposed by the realm page directly.
 
-        :return:
+        :returns: The public key
+        :rtype: str
         """
         params_path = {"realm-name": self.realm_name}
         data_raw = self.connection.raw_get(URL_REALM.format(**params_path))
         return raise_error_from_response(data_raw, KeycloakGetError)["public_key"]
 
     def entitlement(self, token, resource_server_id):
-        """
+        """Get entitlements from the token.
+
         Client applications can use a specific endpoint to obtain a special security token
         called a requesting party token (RPT). This token consists of all the entitlements
         (or permissions) for a user as a result of the evaluation of the permissions and
         authorization policies associated with the resources being requested. With an RPT,
         client applications can gain access to protected resources at the resource server.
 
-        :return:
+        :param token: Access token
+        :type token: str
+        :param resource_server_id: Resource server ID
+        :type resource_server_id: str
+        :returns: Entitlements
+        :rtype: dict
         """
         self.connection.add_param_headers("Authorization", "Bearer " + token)
         params_path = {"realm-name": self.realm_name, "resource-server-id": resource_server_id}
@@ -352,23 +464,28 @@ class KeycloakOpenID:
         if data_raw.status_code == 404:
             return raise_error_from_response(data_raw, KeycloakDeprecationError)
 
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakGetError)  # pragma: no cover
 
     def introspect(self, token, rpt=None, token_type_hint=None):
-        """
+        """Introspect the user token.
+
         The introspection endpoint is used to retrieve the active state of a token.
         It is can only be invoked by confidential clients.
 
         https://tools.ietf.org/html/rfc7662
 
-        :param token:
-        :param rpt:
-        :param token_type_hint:
+        :param token: Access token
+        :type token: str
+        :param rpt: Requesting party token
+        :type rpt: str
+        :param token_type_hint: Token type hint
+        :type token_type_hint: str
 
-        :return:
+        :returns: Token info
+        :rtype: dict
+        :raises KeycloakRPTNotFound: In case of RPT not specified
         """
         params_path = {"realm-name": self.realm_name}
-
         payload = {"client_id": self.client_id, "token": token}
 
         if token_type_hint == "requesting_party_token":
@@ -381,11 +498,11 @@ class KeycloakOpenID:
         payload = self._add_secret_key(payload)
 
         data_raw = self.connection.raw_post(URL_INTROSPECT.format(**params_path), data=payload)
-
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakPostError)
 
     def decode_token(self, token, key, algorithms=["RS256"], **kwargs):
-        """
+        """Decode user token.
+
         A JSON Web Key (JWK) is a JavaScript Object Notation (JSON) data
         structure that represents a cryptographic key.  This specification
         also defines a JWK Set JSON data structure that represents a set of
@@ -395,34 +512,44 @@ class KeycloakOpenID:
 
         https://tools.ietf.org/html/rfc7517
 
-        :param token:
-        :param key:
-        :param algorithms:
-        :return:
+        :param token: Keycloak token
+        :type token: str
+        :param key: Decode key
+        :type key: str
+        :param algorithms: Algorithms to use for decoding
+        :type algorithms: list[str]
+        :param kwargs: Keyword arguments
+        :type kwargs: dict
+        :returns: Decoded token
+        :rtype: dict
         """
-
         return jwt.decode(token, key, algorithms=algorithms, audience=self.client_id, **kwargs)
 
     def load_authorization_config(self, path):
-        """
-        Load Keycloak settings (authorization)
+        """Load Keycloak settings (authorization).
 
         :param path: settings file (json)
-        :return:
+        :type path: str
         """
-        authorization_file = open(path, "r")
-        authorization_json = json.loads(authorization_file.read())
+        with open(path, "r") as fp:
+            authorization_json = json.load(fp)
+
         self.authorization.load_config(authorization_json)
-        authorization_file.close()
 
     def get_policies(self, token, method_token_info="introspect", **kwargs):
-        """
-        Get policies by user token
+        """Get policies by user token.
 
-        :param token: user token
-        :return: policies list
+        :param token: User token
+        :type token: str
+        :param method_token_info: Method for token info decoding
+        :type method_token_info: str
+        :param kwargs: Additional keyword arguments
+        :type kwargs: dict
+        :return: Policies
+        :rtype: dict
+        :raises KeycloakAuthorizationConfigError: In case of bad authorization configuration
+        :raises KeycloakInvalidTokenError: In case of bad token
         """
-
         if not self.authorization.policies:
             raise KeycloakAuthorizationConfigError(
                 "Keycloak settings not found. Load Authorization Keycloak settings."
@@ -448,15 +575,19 @@ class KeycloakOpenID:
         return list(set(policies))
 
     def get_permissions(self, token, method_token_info="introspect", **kwargs):
-        """
-        Get permission by user token
+        """Get permission by user token.
 
         :param token: user token
+        :type token: str
         :param method_token_info: Decode token method
+        :type method_token_info: str
         :param kwargs: parameters for decode
-        :return: permissions list
+        :type kwargs: dict
+        :returns: permissions list
+        :rtype: list
+        :raises KeycloakAuthorizationConfigError: In case of bad authorization configuration
+        :raises KeycloakInvalidTokenError: In case of bad token
         """
-
         if not self.authorization.policies:
             raise KeycloakAuthorizationConfigError(
                 "Keycloak settings not found. Load Authorization Keycloak settings."
@@ -482,8 +613,7 @@ class KeycloakOpenID:
         return list(set(permissions))
 
     def uma_permissions(self, token, permissions=""):
-        """
-        Get UMA permissions by user token with requested permissions
+        """Get UMA permissions by user token with requested permissions.
 
         The token endpoint is used to retrieve UMA permissions from Keycloak. It can only be
         invoked by confidential clients.
@@ -491,10 +621,12 @@ class KeycloakOpenID:
         http://openid.net/specs/openid-connect-core-1_0.html#TokenEndpoint
 
         :param token: user token
+        :type token: str
         :param permissions: list of uma permissions list(resource:scope) requested by the user
-        :return: permissions list
+        :type permissions: str
+        :returns: Keycloak server response
+        :rtype: dict
         """
-
         permission = build_permission_param(permissions)
 
         params_path = {"realm-name": self.realm_name}
@@ -507,22 +639,25 @@ class KeycloakOpenID:
 
         self.connection.add_param_headers("Authorization", "Bearer " + token)
         data_raw = self.connection.raw_post(URL_TOKEN.format(**params_path), data=payload)
-
         return raise_error_from_response(data_raw, KeycloakPostError)
 
     def has_uma_access(self, token, permissions):
-        """
-        Determine whether user has uma permissions with specified user token
+        """Determine whether user has uma permissions with specified user token.
 
         :param token: user token
+        :type token: str
         :param permissions: list of uma permissions (resource:scope)
-        :return: auth status
+        :type permissions: str
+        :return: Authentication status
+        :rtype: AuthStatus
+        :raises KeycloakAuthenticationError: In case of failed authentication
+        :raises KeycloakPostError: In case of failed request to Keycloak
         """
         needed = build_permission_param(permissions)
         try:
             granted = self.uma_permissions(token, permissions)
         except (KeycloakPostError, KeycloakAuthenticationError) as e:
-            if e.response_code == 403:
+            if e.response_code == 403:  # pragma: no cover
                 return AuthStatus(
                     is_logged_in=True, is_authorized=False, missing_permissions=needed
                 )
@@ -538,7 +673,7 @@ class KeycloakOpenID:
             if not scopes:
                 needed.discard(resource)
                 continue
-            for scope in scopes:
+            for scope in scopes:  # pragma: no cover
                 needed.discard("{}#{}".format(resource, scope))
 
         return AuthStatus(
